class FeatureTypes():
    numerical = 'numerical'
    categorical = 'categorical'


REGRESSION = 'regression'
CLASSIFICATION = 'classification'

NUMERICAL = FeatureTypes.numerical
CATEGORICAL = FeatureTypes.categorical
